/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.XExtensionManager;
import org.deckfour.xes.id.XID;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.mapdb.Serializer;

/**
 * Custom {@link Serializer} for {@link XAttributeExternalImpl}. Attributes take
 * up the following amount of disk-space:
 * <p>
 * <li>8 bytes for their ID (including the attribute key) in the BTree;
 * <li>? bytes for BTree overhead;
 * <li>1 byte for their extension and their type;
 * <li>8 bytes (optionally) if they have nested attributes; and
 * <li>? bytes for their actual content.
 * 
 * 
 * @author F. Mannhardt
 * 
 */
final class MapDBAttributeSerializer extends Serializer<ExternalAttribute> {

	private static final int EXTENSIONS_BITMASK = 1 << 6;
	private static final int ATTRIBUTES_BITMASK = 1 << 7;

	private static final int LIST = 8;
	private static final int CONTAINER = 7;
	private static final int ID = 6;
	private static final int TIMESTAMP = 5;
	private static final int LITERAL = 4;
	private static final int DISCRETE = 3;
	private static final int CONTINUOUS = 2;
	private static final int BOOLEAN_FALSE = 1;
	private static final int BOOLEAN_TRUE = 0;

	private final transient Map<XExtension, Integer> extensionMap = new ConcurrentHashMap<>();
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.mapdb.Serializer#serialize(java.io.DataOutput, java.lang.Object)
	 */
	@Override
	public void serialize(DataOutput out, ExternalAttribute a) throws IOException {

		int extensionIndex = getExtensionIndex(a.getExtension());
		boolean hasExtension = extensionIndex != -1;
		boolean hasAttributes = a.hasAttributes();

		if (a instanceof XAttributeBoolean) {
			if (((XAttributeBoolean) a).getValue()) {
				out.writeByte(addMetaData(hasAttributes, hasExtension, BOOLEAN_TRUE));
			} else {
				out.writeByte(addMetaData(hasAttributes, hasExtension, BOOLEAN_FALSE));
			}
		} else if (a instanceof XAttributeContinuous) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, CONTINUOUS));
			out.writeDouble(((XAttributeContinuous) a).getValue());
		} else if (a instanceof XAttributeDiscrete) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, DISCRETE));
			out.writeLong(((XAttributeDiscrete) a).getValue());
		} else if (a instanceof XAttributeTimestamp) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, TIMESTAMP));
			out.writeLong(((XAttributeTimestamp) a).getValueMillis());
		} else if (a instanceof XAttributeID) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, ID));
			XID.write(((XAttributeID) a).getValue(), out);
		} else if (a instanceof XAttributeContainerExternalImpl) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, CONTAINER));
		} else if (a instanceof XAttributeListExternalImpl) {
			out.writeByte(addMetaData(hasAttributes, hasExtension, LIST));
			out.writeInt(((XAttributeListExternalImpl) a).getKeyOrder().size());
			for (String key : ((XAttributeListExternalImpl) a).getKeyOrder()) {
				out.writeUTF(key);
			}
		} else if (a instanceof XAttributeLiteral) {
			// Literal has to be checked after List and Container!!!
			out.writeByte(addMetaData(hasAttributes, hasExtension, LITERAL));
			out.writeUTF(((XAttributeLiteral) a).getValue());
		} else {
			throw new RuntimeException("Serialization failed for attribute " + a);
		}

		if (hasExtension) {
			assert extensionIndex <= Byte.MAX_VALUE : "Too many XExtension registered for MapDB serializer, only supporting 127";
			out.writeByte(extensionIndex);
		}
		if (hasAttributes) {
			out.writeLong(((ExternalIdentifyable) a).getExternalId());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.mapdb.Serializer#deserialize(java.io.DataInput, int)
	 */
	@Override
	public ExternalAttribute deserialize(DataInput in, int available) throws IOException {

		int key = Integer.MIN_VALUE;

		byte type = in.readByte();
		boolean hasAttributes = checkAttributesFlag(type);
		boolean hasExtension = checkExtensionFlag(type);
		type = getRealType(type);

		ExternalAttribute attr = null;
		switch (type) {
			case BOOLEAN_TRUE :
				attr = new XAttributeBooleanExternalImpl(key, true, null);
				break;
			case BOOLEAN_FALSE :
				attr = new XAttributeBooleanExternalImpl(key, false, null);
				break;
			case CONTINUOUS :
				attr = new XAttributeContinuousExternalImpl(key, in.readDouble(), null);
				break;
			case DISCRETE :
				attr = new XAttributeDiscreteExternalImpl(key, in.readLong(), null);
				break;
			case LITERAL :
				attr = new XAttributeLiteralExternalImpl(key, in.readUTF(), null);
				break;
			case TIMESTAMP :
				attr = new XAttributeTimestampExternalImpl(key, in.readLong(), null);
				break;
			case ID :
				XID read = XID.read(in);
				attr = new XAttributeIDExternalImpl(key, read, null);
				break;
			case CONTAINER :
				attr = new XAttributeContainerExternalImpl(key, null, null);
				break;
			case LIST :
				int size = in.readInt();
				List<String> keyOrder = new ArrayList<>(size);
				for (int i = 0; i < size; i++) {
					keyOrder.add(in.readUTF());
				}
				attr = new XAttributeListExternalImpl(key, keyOrder, null, null);
				break;
			default :
				throw new RuntimeException("De-Serialization failed for attribute with key " + key);
		}

		if (hasExtension) {
			byte ext = in.readByte();
			attr.setExtension(XExtensionManager.instance().getByIndex(ext));
		}

		if (hasAttributes) {
			long metaAttributesId = in.readLong();
			attr.setExternalId(metaAttributesId);
		}
		return attr;
	}

	private int getExtensionIndex(XExtension extension) {
		if (extension == null) {
			return -1;
		} else {
			Integer index = getExtensionMap().get(extension);
			if (index != null) {
				return index.intValue();
			} else {
				int index2 = XExtensionManager.instance().getIndex(extension);
				getExtensionMap().put(extension, index2);
				return index2;
			}			
		}
	}

	private static byte addMetaData(boolean hasAttributes, boolean hasExtension, int typeInt) {
		byte type = (byte) typeInt;
		assert (type & (ATTRIBUTES_BITMASK | EXTENSIONS_BITMASK)) == 0 : "Type in XAttributeSerializer should not be negative!";
		if (hasAttributes) {
			type |= ATTRIBUTES_BITMASK;
		}
		if (hasExtension) {
			type |= EXTENSIONS_BITMASK;
		}
		return type;
	}

	private static byte getRealType(byte type) {
		// Reset the meta data bits
		type &= ~(ATTRIBUTES_BITMASK | EXTENSIONS_BITMASK);
		return type;
	}

	private static boolean checkAttributesFlag(byte type) {
		// Check if highest bit is set
		return (byte) (type & ATTRIBUTES_BITMASK) != 0;
	}

	private static boolean checkExtensionFlag(byte type) {
		// Check if highest bit is set
		return (byte) (type & EXTENSIONS_BITMASK) != 0;
	}

	private Map<XExtension, Integer> getExtensionMap() {
		return extensionMap;
	}

}