package org.processmining.xeslite.external;

import java.nio.ByteBuffer;
import java.util.AbstractList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.NoSuchElementException;

import net.jpountz.lz4.LZ4Factory;

import com.google.common.primitives.Longs;

abstract class XAbstractCompressedList<E> extends AbstractList<E> {

	interface Compressor {

		byte[] decompress(byte[] compressedData, int uncompressedSize);

		byte[] compress(byte[] array);

	}

	static final class EventData {
		long[] ids;
		ByteBuffer[] data;

		EventData() {
			this(0);
		}

		EventData(int size) {
			super();
			ids = new long[size];
			data = new ByteBuffer[size];
		}
	}

	//TODO these iterators are not fast failing!
	private class CompressedIterator implements Iterator<E> {

		protected EventData eventData;

		protected int index = 0;
		protected int last = -1;

		CompressedIterator(EventData eventData) {
			this.eventData = eventData;
		}

		public boolean hasNext() {
			return index != size();
		}

		public E next() {
			if (!hasNext()) {
				throw new NoSuchElementException();
			}
			E element = newInstance(index, eventData.ids[index], eventData.data[index]);
			last = index;
			index++;
			return element;
		}

		public void remove() {
			if (last < 0) {
				throw new IllegalStateException();
			}

			//TODO update avoid re-reading upon each remove
			// Remove from list			
			XAbstractCompressedList.this.remove(last);
			// Update cached event data
			eventData = getEventData();
			
			if (last < index) {
				index--;
			}
			last = -1;
		}

	}

	//TODO these iterators are not fast failing!
	private class CompressedListIterator extends CompressedIterator implements ListIterator<E> {

		CompressedListIterator(int index, EventData eventData) {
			super(eventData);
			this.index = index;
		}

		public boolean hasPrevious() {
			return index > 0;
		}

		public E previous() {
			try {
				int prevIndex = index - 1;
				E previous = get(prevIndex);
				index = prevIndex;
				last = prevIndex;
				return previous;
			} catch (IndexOutOfBoundsException e) {
				throw new NoSuchElementException();
			}
		}

		public int nextIndex() {
			return index;
		}

		public int previousIndex() {
			return index - 1;
		}

		public void set(E e) {
			if (last == -1) {
				throw new IllegalStateException();
			}
			XAbstractCompressedList.this.set(last, e);
		}

		public void add(E e) {
			XAbstractCompressedList.this.add(index++, e);
			last = -1;
		}

	}

	private static final Compressor COMPRESSOR = new Compressor() {
		
		private final LZ4Factory factory = LZ4Factory.fastestJavaInstance();

		public byte[] decompress(byte[] compressedData, int uncompressedSize) {
			byte[] uncompressedBuffer = new byte[uncompressedSize];
			factory.fastDecompressor().decompress(compressedData, uncompressedBuffer, uncompressedSize);
			return uncompressedBuffer;
		}

		public byte[] compress(byte[] data) {
			return factory.fastCompressor().compress(data);
		}

	};

	private final int payloadSize;

	private byte[] compressedData;
	private int size;

	public XAbstractCompressedList(int payloadSize) {
		this.payloadSize = payloadSize;
		this.compressedData = null;
		this.size = 0;
	}

	abstract protected E newInstance(int index, long id, ByteBuffer data);

	abstract protected E convertElement(E e);

	abstract protected long getExternalId(E e);
	
	abstract protected int getIdShift();

	abstract protected ByteBuffer convertData(E e);
	
	protected ByteBuffer getCompressedIds() {
		return ByteBuffer.wrap(compressedData);
	}

	public E get(int index) {
		if (index < 0 || index >= size()) {
			throw new IndexOutOfBoundsException();
		}
		return newInstance(index, getEventData().ids[index], getEventData().data[index]);
	}

	public E set(int index, E element) {
		if (index < 0 || index >= size()) {
			throw new IndexOutOfBoundsException();
		}
		E newElement = convertElement(element);

		EventData eventData = getEventData();

		E oldElement = newInstance(index, eventData.ids[index], eventData.data[index]);

		eventData.ids[index] = getExternalId(newElement);
		eventData.data[index] = convertData(newElement);

		setEventData(eventData);

		return oldElement;
	}

	public void add(int index, E element) {
		if (index < 0 || index > size()) {
			throw new IndexOutOfBoundsException();
		}
		E newElement = convertElement(element);
		EventData eventData = resizeIfNeeded(size + 1, getEventData());

		System.arraycopy(eventData.ids, index, eventData.ids, index + 1, size - index);
		System.arraycopy(eventData.data, index, eventData.data, index + 1, size - index);

		eventData.ids[index] = getExternalId(newElement);
		eventData.data[index] = convertData(newElement);
		size++;

		setEventData(eventData);
	}

	public boolean addAll(Collection<? extends E> c) {
		return addAll(size, c);
	}

	public boolean addAll(int index, Collection<? extends E> c) {
		if (index < 0 || index > size()) {
			throw new IndexOutOfBoundsException();
		}
		int collectionSize = c.size();

		EventData eventData = resizeIfNeeded(size + collectionSize, getEventData());

		System.arraycopy(eventData.ids, index, eventData.ids, index + collectionSize, size - index);
		System.arraycopy(eventData.data, index, eventData.data, index + collectionSize, size - index);

		int i = 0;
		for (Iterator<? extends E> iterator = c.iterator(); iterator.hasNext(); i++) {
			E element = iterator.next();
			E newElement = convertElement(element);
			eventData.ids[index + i] = getExternalId(newElement);
			eventData.data[index + i] = convertData(newElement);
		}
		size += collectionSize;

		setEventData(eventData);

		return collectionSize != 0;
	}

	public Iterator<E> iterator() {
		return new CompressedIterator(getEventData());
	}

	public ListIterator<E> listIterator(int index) {
		return new CompressedListIterator(index, getEventData());
	}

	public E remove(int index) {
		if (index < 0 || index >= size()) {
			throw new IndexOutOfBoundsException();
		}

		EventData eventData = getEventData();

		E oldEvent = newInstance(index, eventData.ids[index], eventData.data[index]);

		eventData = doRemove(index, eventData);

		setEventData(eventData);

		return oldEvent;
	}

	private EventData doRemove(int index, EventData eventData) {
		// Move subsequent events
		int tailLength = size - index - 1;
		if (tailLength > 0) {
			System.arraycopy(eventData.ids, index + 1, eventData.ids, index, tailLength);
			System.arraycopy(eventData.data, index + 1, eventData.data, index, tailLength);
		}

		// Change size and set last entry to zero
		--size;
		eventData.ids[size] = 0;
		eventData.data[size] = null;
		
		return eventData;
	}

	public final int size() {
		return size;
	}

	protected final EventData getEventData() {
		if (compressedData == null) {
			return new EventData();
		} else {
			int uncompressedSize = uncompressedSize();
			byte[] uncompressedData = COMPRESSOR.decompress(compressedData, uncompressedSize);
			return decodeFromBytes(ByteBuffer.wrap(uncompressedData));
		}
	}

	protected final void setEventData(EventData eventIds) {
		byte[] uncompressedData = encodeToBytes(eventIds).array();
		compressedData = COMPRESSOR.compress(uncompressedData);
	}

	private final ByteBuffer encodeToBytes(EventData eventData) {
		ByteBuffer buffer = ByteBuffer.allocate(uncompressedSize());
		
		long lastId = 0;
		for (int i = 0; i < size; i++) {
			long id = eventData.ids[i] >> getIdShift();

			// Delta encode
			long delta = id - lastId;
			lastId = id;

			buffer.putLong(delta);
			if (hasPayload()) {
				assert eventData.data[i].remaining() == payloadSize;
				buffer.put(eventData.data[i]);
				// Reset original byte buffer
				eventData.data[i].position(0);
			}
		}
		buffer.flip();
		return buffer;
	}

	protected final boolean hasPayload() {
		return payloadSize != 0;
	}

	private final int uncompressedSize() {
		return size * (Longs.BYTES + payloadSize);	
	}

	private final EventData decodeFromBytes(ByteBuffer uncompressedData) {
		EventData eventData = new EventData(size);

		long lastId = 0;

		for (int i = 0; i < size; i++) {

			// Delta decode
			long newId = uncompressedData.getLong() + lastId;
			eventData.ids[i] = newId << getIdShift();
			lastId = newId;

			if (hasPayload()) {
				// Use shared array
				ByteBuffer payload = uncompressedData.slice();
				payload.limit(payload.position() + payloadSize);
				eventData.data[i] = payload;
				uncompressedData.position(uncompressedData.position() + payloadSize);				
			}
			
		}

		return eventData;
	}

	private static EventData resizeIfNeeded(int requiredCapacity, EventData data) {
		data.ids = resizeIfNeeded(requiredCapacity, data.ids);
		data.data = resizeIfNeeded(requiredCapacity, data.data);
		return data;
	}

	private static long[] resizeIfNeeded(int requiredCapacity, long[] array) {
		if (requiredCapacity > array.length) {
			int newCapacity = Math.max(10, 2 * array.length);
			if (newCapacity < requiredCapacity) {
				newCapacity = requiredCapacity;
			}
			array = Arrays.copyOf(array, newCapacity);
		}
		return array;
	}

	private static ByteBuffer[] resizeIfNeeded(int requiredCapacity, ByteBuffer[] array) {
		if (requiredCapacity > array.length) {
			int newCapacity = Math.max(10, 2 * array.length);
			if (newCapacity < requiredCapacity) {
				newCapacity = requiredCapacity;
			}
			array = Arrays.copyOf(array, newCapacity);
		}
		return array;
	}

}
